from ast import Raise
from curses.ascii import HT
from diogi.functions import *

class Doc:
    #href, format, tag
    def __init__(self, href: str=None, format: str=None, tag: str=None, value: str=None):
         self.href = href 
         self.format = format
         self.tag = tag 
         self.value = value 

    def as_data(self):
        data = {}
        set_if_not_none(data, self.href, 'href')
        set_if_not_none(data, self.format, 'format')
        set_if_not_none(data, self.tag, 'tag')
        set_if_not_none(data, self.value, 'value')
        return data 

    def parse(obj: any):
        if str == type(obj):
            return TextDoc(obj)

        if dict == type(obj):
            value = get_if_exists(obj, 'value', None)
            format = get_if_exists(obj, 'format', 'text')
            href = get_if_exists(obj, 'href', None)
            tag = get_if_exists(obj, 'tag', None)

            if format == 'markdown':
                return MarkDownDoc(
                    value=default_if_none(value, ''), 
                    href=href, 
                    tag=tag
                )
            if format == 'html':
                return HtmlDoc(
                    href = href,
                    tag = tag,
                    value = default_if_none(value, '')
                )
                
            return TextDoc(
                value=default_if_none(value, str(obj)), 
                href=href, 
                tag=tag
            )

        return TextDoc()

    def __eq__(self, other):
        if type(other) is type(self):
            return self.format == other.format and \
                self.href == other.href and \
                self.tag == other.tag and \
                self.value == other.value 
        else:
            return False

    def __hash__(self):
        return hash((self.format, self.href, self.tag, self.value))


class WithDocsMixin:
    def __init__(self, *args, **kwargs):
        super(WithDocsMixin, self).__init__(*args, **kwargs)
    
    def add_doc(self, doc: any):
        if not isinstance(doc, Doc):
            doc = Doc.parse(doc)
        append_if_not_none(self.contents, doc, 'doc')
        return self

    @property
    def docs(self):
        return always_a_list(get_if_exists(self.contents, 'doc', []))
        

class HtmlDoc(Doc):
    def __init__(self, href: str, value: str, *args, **kwargs):
        super(HtmlDoc, self).__init__(
            href = href,
            format = 'html',
            value = value,
            *args, **kwargs
        )


class MarkDownDoc(Doc):
    def __init__(self, value: str, *args, **kwargs):
        super(MarkDownDoc, self).__init__(
            format = 'markdown',
            value = value,
            *args, **kwargs
        )


class TextDoc(Doc):
    def __init__(self, value: str='', *args, **kwargs):
        super(TextDoc, self).__init__(
            format = 'text',
            value=value, 
            *args, **kwargs
        )


import logging 
from .descriptors import DescriptorBase
from diogi.conventions import to_data
from diogi.functions import get_if_exists, always_a_list, default_if_none

logger = logging.getLogger(__name__)

class Alps(DescriptorBase):
    """
    Root ALPS Descriptor
    """
    def __init__(self, version: str='1.0', *args, **kwargs):
        super(Alps, self).__init__(*args, **kwargs)
        self.contents['version'] = version
        self.contents['title'] = None
        
    @property
    def version(self):
        return self.contents['version']

    @property 
    def title(self):
        return self.contents['title']

    def set_title(self, title: str):
        self.contents['title'] = title

    def as_data(self):
        return { 'alps': to_data(super()) }

    def default_resolver(data: dict):
        def resolver(href: str):
            if str != type(href) or len(href) < 2: 
                # TODO: warn in logs? 
                return {}
        
            if href[0:1] == '#':
                descriptors = always_a_list(get_if_exists(data['alps'], 'descriptor', []))
                matches = [d for d in descriptors if d['id'] == href[1::]]
                if len(matches)>0:
                    return matches[0]
                return {}
                
        return resolver 


    def parse(obj: dict, href_resolver: callable): 
        if obj is None or \
            not dict == type(obj) or \
            not 'alps' in obj.keys():
            logger.warning('The document provided does not contain ALPS definition!')
            return Alps(version=None)

        data = obj['alps']
        version = get_if_exists(data, 'version', '1.0')
        if (version != '1.0'): 
            logger.warning(f'The document is in unsupported ALPS version {version}')

        alps = Alps(version=version)
        alps.source = obj
        alps.set_title(get_if_exists(data, 'title', None))

        for d in always_a_list(get_if_exists(data, 'doc', [])):
            alps.add_doc(d)

        for d in always_a_list(get_if_exists(data, 'descriptor', [])):
            alps.add_descriptor(d, href_resolver)

        return alps


from diogi.functions import *
from diogi.conventions import to_data
from .docs import WithDocsMixin

def noop_resolver(href: str) -> dict: 
    pass 

class Descriptor: 
    def parse(obj: any, resolver: callable):
        if dict == type(obj):
            href = get_if_exists(obj, 'href', None)
            if href: 
                obj = {**default_if_none(resolver(href), {}), **obj}
            desc_type = get_if_exists(obj, 'type', 'semantic')
            docs = get_if_exists(obj, 'doc', None)

        else: 
            return None

        desc = None 

        if desc_type == 'semantic':
            desc = Semantic(
                id=get_if_exists(obj, 'id'),
                name=get_if_exists(obj, 'name')
            )

        if desc_type == 'safe':
            desc = Safe(
                id=get_if_exists(obj, 'id'), 
                name=get_if_exists(obj, 'name')
            )

        if docs: 
            add_doc = getattr(desc, 'add_doc', None) 
            if add_doc: 
                add_doc(docs)

        for d in always_a_list(get_if_exists(obj, 'descriptor', [])):
            desc.add_descriptor(d, resolver)

        return desc
 

class DescriptorBase(WithDocsMixin):
    def __init__(self):
        self.contents = {}

    @property
    def id(self):
        return get_if_exists(self.contents, 'id', None)

    @property 
    def name(self):
        return get_if_exists(self.contents, 'name', None)

    @property 
    def descriptors(self):
        return always_a_list(get_if_exists(self.contents, 'descriptor', []))

    def add_descriptor(self, 
        descriptor: Descriptor,
        resolver: callable = noop_resolver
    ):
        if not isinstance(descriptor, Descriptor):
            descriptor = Descriptor.parse(descriptor, resolver)

        append_if_not_none(self.contents, descriptor, 'descriptor')
        return self

    def get_descriptor(self, id: str) -> Descriptor:
        return list_is_optional([d for d in get_if_exists(self.contents, 'descriptor', []) if d.id == id])

    def as_data(self): 
        data = {}
        for k,v in self.contents.items():
            set_if_not_none(data, to_data(list_is_optional(v)), k)

        return data

    def __eq__(self, other):
        if type(other) is type(self):
            return self.contents == other.contents
        else:
            return False

    def __hash__(self):
        return hash((self.contents, self.contents))



class SimpleDescriptor(Descriptor, DescriptorBase):
    def __init__(self, 
        id: str=None, 
        text: str=None, 
        ref: str=None, 
        name: str=None,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.contents['id'] = id
        self.contents['text'] = text 
        self.contents['ref'] = ref 
        self.contents['name'] = name


class Idempotent():
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.contents['type'] = 'idempotent'


class ReferencingDescriptor(SimpleDescriptor):
    def __init__(self, ref: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.contents['ref'] = ref


class Safe(SimpleDescriptor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.contents['type'] = 'safe'    


class Semantic(SimpleDescriptor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.contents['type'] = 'semantic'


class Unsafe(SimpleDescriptor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.contents['type'] = 'unsafe'
